from . import clients
from . import search
from . import models
from . import utils
from . import exceptions

Client = clients.Client
logger = utils.MagentoLogger(
    name=utils.MagentoLogger.PACKAGE_LOG_NAME,
    log_file=utils.MagentoLogger.PACKAGE_LOG_NAME + '.log',
    stdout_level='WARNING'  # Clients will log to console
)
logger.debug('Initialized MyMagento')



from __future__ import annotations
from functools import cached_property
from typing import Union, Type, Iterable, List, Optional, Dict, TYPE_CHECKING
from .models import Model, APIResponse, Product, Category, ProductAttribute, Order, OrderItem, Invoice
from .exceptions import MagentoError
from . import clients

if TYPE_CHECKING:
    from typing_extensions import Self
    from . import Client


class SearchQuery:

    """Queries any endpoint that invokes the searchCriteria interface. Parent of all endpoint-specific search classes

        .. tip:: See https://developer.adobe.com/commerce/webapi/rest/use-rest/performing-searches/ for official docs
    """

    def __init__(self, endpoint: str, client: Client, model: Type[Model] = APIResponse):
        """Initialize a SearchQuery object

        :param endpoint: the base search API endpoint (for example, ``orders``)
        :param client: an initialized :class:`~.Client` object
        :param model: the :class:`~.Model` to parse the response data with; uses :class:`~.APIResponse` if not specified
        """
        if not isinstance(client, clients.Client):
            raise TypeError(f'`client` must be of type {clients.Client}')

        #: The :class:`~.Client` to send the search request with
        self.client = client
        #: The endpoint being queried
        self.endpoint = endpoint
        #: The :doc:`Model <models>`_ class to wrap the response with
        self.Model = model
        #: The current url for the search request
        self.query = self.client.url_for(endpoint) + '/?'
        #: Restricted fields, from :meth:`~restricted_fields`
        self.fields = ''
        #: The raw response data, if any
        self._result = {}

    def add_criteria(self, field, value, condition='eq', **kwargs) -> Self:
        """Add criteria to the search query

        :param field: the API response field to search by
        :param value: the value of the field to compare
        :param condition: the comparison condition
        :param kwargs: additional search option arguments (``group`` and ``filter``)
        :returns: the calling SearchQuery object

        *Options*
                condition:  condition used to evaluate the attribute value

                            (Default Value)     'eq'            =           (field=value)
                                                'gt'            >
                                                'lt'            <
                                                'gteq'          >=
                                                'lteq'          <=
                                                'in'            []


                group:      filter group number

                filter:     filter number (within the specified filter group)


        *Using Filter Groups*

            Filter groups are filter criteria in the form of { field: value }

                Group 0 Filter 0                        ->      Filter 0
                Group 0 Filter 0 + Group 0 Filter 1     ->      Filter 0 OR Filter 1
                Group 0 Filter 0 + Group 1 Filter 0     ->      Filter 0 AND Filter 0
        """

        options = {
            'condition': condition,
            'group': 0,
            'filter': 0,
        }
        options.update(kwargs)

        criteria = (
                f'searchCriteria[filter_groups][{options["group"]}][filters][{options["filter"]}][field]={field}' +
                f'&searchCriteria[filter_groups][{options["group"]}][filters][{options["filter"]}][value]={value}' +
                f'&searchCriteria[filter_groups][{options["group"]}][filters][{options["filter"]}][condition_type]={options["condition"]}'
        )
        if not self.query.endswith('?'):
            self.query += '&'
        self.query += criteria
        return self

    def restrict_fields(self, fields: Iterable[str]) -> Self:
        """Constrain the API response data to only contain the specified fields

        :param fields: an iterable or comma separated string of fields to include in the response
        :returns: the calling SearchQuery object
        """
        if not isinstance(fields, str):
            if not isinstance(fields, Iterable):
                raise TypeError('"fields" must be a comma separated string or an iterable')
            fields = ','.join(fields)

        if (id_field := self.Model.IDENTIFIER) not in fields:
            fields += f',{id_field}'

        self.fields = f'&fields=items[{fields}]'
        return self

    def execute(self) -> Optional[Model | List[Model]]:
        """Sends the search request using the current :attr:`~.scope` of the :attr:`client`

        .. tip:: Change the :attr:`.Client.scope` to retrieve :attr:`~.result` data
           from different store :attr:`~.views`

        :returns: the search query :attr:`~.result`
        """
        response = self.client.get(self.query + self.fields)
        self.__dict__.pop('result', None)
        self._result = response.json()
        return self.result

    def by_id(self, item_id: Union[int, str]) -> Optional[Model]:
        """Retrieve data for an individual item by its id

        .. note:: The ``id`` field used is different depending on the endpoint being queried

           * Most endpoints use an ``entity_id`` or ``id``
           * The ``orders/items`` endpoint uses ``item_id``
           * The ``products`` endpoint uses ``product_id``,
             but can also be queried :meth:`~ProductSearch.by_sku`

           The :attr:`~.Model.IDENTIFIER` attribute of each :class:`~.Model` contains the appropriate field

        :param item_id: id of the item to retrieve
        """
        self.query = self.query.strip('?') + str(item_id)
        return self.execute()

    def by_list(self, field: str, values: Iterable) -> Optional[Model, List[Model]]:
        """Search for multiple items using an iterable or comma-separated string of field values

        .. admonition:: Example
           :class: example

           Search for orders that are processing, pending, or completed::

            #  Values can be a list/tuple/iterable
            >> api.orders.by_list('status', ['processing', 'pending', 'completed'])

            #  Values can be a comma-separated string
            >> api.orders.by_list('status', 'processing,pending,completed')


        :param field: the API response field to search for matches in
        :param values: an iterable or comma separated string of values
        """
        if not isinstance(values, Iterable):
            raise TypeError('`values` must be an iterable')
        if not isinstance(values, str):
            values = ','.join(f'{value}' for value in values)
        return self.add_criteria(
            field=field,
            value=values,
            condition='in'
        ).execute()

    @cached_property
    def result(self) -> Optional[Model | List[Model]]:
        """The result of the search query, wrapped by the :class:`~.Model` corresponding to the endpoint

        :returns: the API response as either an individual or list of :class:`~.Model` objects
        """
        result = self.validate_result()
        if result is None:
            return result
        if isinstance(result, list):
            return [self.parse(item) for item in result]
        if isinstance(result, dict):
            return self.parse(result)

    def validate_result(self) -> Optional[Dict | List[Dict]]:
        """Parses the response and returns the actual result data, regardless of search approach"""
        if not self._result:
            return None

        if isinstance(self._result, list):
            return self._result

        if self._result.get('message'):  # Error; logged by Client
            return None

        if len(self._result.keys()) > 3:  # Single item, retrieved by id
            return self._result

        if 'items' in self._result:  # All successful responses with search criteria have `items` key
            items = self._result['items']
            if items:  # Response can still be {'items': None} though
                if len(items) > 1:
                    return items
                return items[0]
            else:
                self.client.logger.info(
                    "No matching {} for this search query".format(self.endpoint)
                )
                return None
        # I have no idea what could've gone wrong, sorry :/
        msg = "Search failed with an unknown error.\nAPI Response: {}".format(self._result)
        raise MagentoError(self.client, msg)

    def parse(self, data: dict) -> Model:
        """Parses the API response with the corresponding :class:`~.Model` object

        :param data: API response data of a single item
        """
        if self.Model is not APIResponse:
            return self.Model(data, self.client)
        return self.Model(data, self.client, self.endpoint)

    def reset(self) -> None:
        """Resets the query and result, allowing the object to be reused"""
        self._result = {}
        self.fields = ''
        self.query = self.client.url_for(self.endpoint) + '/?'
        self.__dict__.pop('result', None)

    @property
    def result_count(self) -> int:
        """Number of items that matched the search criteria"""
        if not self._result or not self.result:
            return 0
        if isinstance(self.result, Model):
            return 1
        return len(self.result)

    @property
    def result_type(self) -> Type:
        """The type of the result"""
        return type(self.result)


class OrderSearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``orders`` endpoint"""

    def __init__(self, client: Client):
        """Initialize an :class:`OrderSearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='orders',
            client=client,
            model=Order
        )

    def by_number(self, order_number: Union[int, str]) -> Optional[Order]:
        """Retrieve an :class:`~.Order` by number

        :param order_number: the order number (``increment_id``)
        """
        return self.add_criteria(
            field='increment_id',
            value=order_number
        ).execute()

    def by_product(self, product: Product) -> Optional[Order | List[Order]]:
        """Search for all :class:`~.Order` s of a :class:`~.Product`

        :param product: the :class:`~.Product` to search for in orders
        """
        items = self.client.order_items.by_product(product)
        return self.from_items(items)

    def by_sku(self, sku: str) -> Optional[Order | List[Order]]:
        """Search for :class:`~.Order` by product sku

        .. note:: Like :meth:`.OrderItemSearch.by_sku`, the sku will need to be an exact
           match to the sku of a simple product, including a custom option if applicable

           * Use :meth:`~.OrderSearch.by_product` or :meth:`~.OrderSearch.by_product_id`
             to find orders containing any of the :attr:`~.option_skus` and/or all
             :attr:`~.children` of a configurable product

        :param sku: the exact product sku to search for in orders
        """
        items = self.client.order_items.by_sku(sku)
        return self.from_items(items)

    def by_product_id(self, product_id: Union[int, str]) -> Optional[Order | List[Order]]:
        """Search for :class:`~.Order` s by ``product_id``

        :param product_id: the ``id`` (``product_id``) of the product to search for in orders
        """
        items = self.client.order_items.by_product_id(product_id)
        return self.from_items(items)

    def by_category_id(self, category_id: Union[int, str], search_subcategories: bool = False) -> Optional[Order | List[Order]]:
        """Search for :class:`~.Order` s by ``category_id``

        :param category_id: id of the category to search for in orders
        :param search_subcategories: if ``True``, also searches for orders from :attr:`~.all_subcategories`
        :returns: any :class:`~.Order` containing a :class:`~.Product` in the corresponding :class:`~.Category`
        """
        items = self.client.order_items.by_category_id(category_id, search_subcategories)
        return self.from_items(items)

    def by_category(self, category: Category, search_subcategories: bool = False) -> Optional[Order | List[Order]]:
        """Search for :class:`~.Order` s that contain any of the category's :attr:`~.Category.products`

        :param category: the :class:`~.Category` to use in the search
        :param search_subcategories: if ``True``, also searches for orders from :attr:`~.all_subcategories`
        :returns: any :class:`~.Order` that contains a product in the provided category
        """
        items = self.client.order_items.by_category(category, search_subcategories)
        return self.from_items(items)

    def by_skulist(self, skulist: Union[str, Iterable[str]]) -> Optional[Order | List[Order]]:
        """Search for :class:`~.Order` s using a list or comma separated string of product SKUs

        :param skulist: an iterable or comma separated string of product SKUs
        """
        items = self.client.order_items.by_skulist(skulist)
        return self.from_items(items)

    def from_items(self, items: Optional[OrderItem | List[OrderItem]]) -> Optional[Order, List[Order]]:
        """Retrieve unique :class:`~.Order` objects from :class:`~.OrderItem` entries using a single request

        :param items: an individual/list of order items
        """
        if items is None:
            return
        if isinstance(items, list):
            order_ids = set(item.order_id for item in items)
            return self.by_list('entity_id', order_ids)
        else:
            return items.order  # Single OrderItem


class OrderItemSearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``orders/items`` endpoint"""

    def __init__(self, client: Client):
        """Initialize an :class:`OrderItemSearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='orders/items',
            client=client,
            model=OrderItem
        )

    @cached_property
    def result(self) -> Optional[OrderItem | List[OrderItem]]:
        if result := super().result:
            if isinstance(result, list):
                return [item for item in result if item]
        return result

    def parse(self, data) -> Optional[OrderItem]:
        """Overrides :meth:`SearchQuery.parse` to fully hydrate :class:`~.OrderItem` objects

        Extra validation is required for OrderItems, as duplicated and/or incomplete data is returned
        when the child of a configurable product is searched :meth:`by_sku` or :meth:`by_product`

        :param data: API response data
        """
        if data.get('parent_item'):
            return None
        if parent_id := data.get('parent_item_id'):
            return self.client.order_items.by_id(parent_id)
        else:
            return OrderItem(data, self.client)

    def by_product(self, product: Product) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem` entries by :class:`~.Product`

        .. note:: This will match OrderItems that contain

           * Any of the child products of a configurable product
           * Any of the :attr:`~.option_skus` of a product with custom options

        :param product: the :class:`~.Product` to search for in order items
        """
        if not isinstance(product, Product):
            raise TypeError(f'`product` must be of type {Product}')

        return self.by_product_id(product.id)

    def by_sku(self, sku: str) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem` entries by product sku.

        .. admonition:: The SKU must be an exact match to the OrderItem SKU

           OrderItems always use the SKU of a simple product, including any custom options.
           This means that:

           * Searching the SKU of a configurable product returns nothing
           * If a product has custom options, the search will only find OrderItems
             that contain the specific option sku (or base sku) that's provided

           To search for OrderItems containing all :attr:`~.children` of a
           configurable product and/or all possible :attr:`~.option_skus`,
           use :meth:`~.by_product` or :meth:`~.by_product_id`

        :param sku: the exact product sku to search for in order items
        """
        return self.add_criteria('sku', Model.encode(sku)).execute()

    def by_product_id(self, product_id: Union[int, str]) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem` entries by product id.

        :param product_id: the ``id`` (``product_id``) of the :class:`~.Product` to search for in order items
        """
        return self.add_criteria('product_id', product_id).execute()

    def by_category_id(self, category_id: Union[int, str], search_subcategories: bool = False) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem` entries by ``category_id``

        :param category_id: id of the :class:`~.Category` to search for in order items
        :param search_subcategories: if ``True``, also searches for order items from :attr:`~.all_subcategories`
        :returns: any :class:`~.OrderItem` containing a :class:`~.Product` in the corresponding :class:`~.Category`
        """
        if category := self.client.categories.by_id(category_id):
            return self.by_category(category, search_subcategories)

    def by_category(self, category: Category, search_subcategories: bool = False) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem` entries that contain any of the category's :attr:`~.Category.products`

        :param category: the :class:`~.Category` to use in the search
        :param search_subcategories: if ``True``, also searches for order items from :attr:`~.all_subcategories`
        """
        if not isinstance(category, Category):
            raise TypeError(f'`category` must be of type {Category}')

        product_ids = category.all_product_ids if search_subcategories else category.product_ids
        return self.by_list('product_id', product_ids)

    def by_skulist(self, skulist: Union[str, Iterable[str]]) -> Optional[OrderItem | List[OrderItem]]:
        """Search for :class:`~.OrderItem`s using a list or comma-separated string of product SKUs

        :param skulist: an iterable or comma separated string of product SKUs
        """
        if not isinstance(skulist, Iterable):
            raise TypeError(f'`skulist` must be an iterable or comma-separated string of SKUs')
        if isinstance(skulist, str):
            skulist = skulist.split(',')

        skus = map(Model.encode, skulist)
        return self.by_list('sku', skus)


class InvoiceSearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``invoices`` endpoint"""

    def __init__(self, client: Client):
        """Initialize an :class:`InvoiceSearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='invoices',
            client=client,
            model=Invoice
        )

    def by_number(self, invoice_number: Union[int, str]) -> Optional[Invoice]:
        """Retrieve an :class:`~.Invoice` by number

        :param invoice_number: the invoice number (``increment_id``)
        """
        return self.add_criteria(
            field='increment_id',
            value=invoice_number
        ).execute()

    def by_order_number(self, order_number: Union[int, str]) -> Optional[Invoice]:
        """Retrieve an :class:`~.Invoice` by order number

        :param order_number: the order number (``increment_id``)
        """
        if order := self.client.orders.by_number(order_number):
            return self.by_order(order)

    def by_order(self, order: Order) -> Optional[Invoice]:
        """Retrieve the :class:`~.Invoice` for an :class:`~.Order`

        :param order: the :class:`~.Order` object to retrieve an invoice for
        """
        return self.by_order_id(order.id)

    def by_order_id(self, order_id: Union[int, str]) -> Optional[Invoice]:
        """Retrieve an :class:`~.Invoice` by ``order_id``

        :param order_id: the ``order_id`` of the order to retrieve an invoice for
        """
        return self.add_criteria(
            field='order_id',
            value=order_id
        ).execute()

    def by_product(self, product: Product) -> Optional[Order | List[Order]]:
        """Search for all :class:`~.Invoice` s of a :class:`~.Product`

        :param product: the :class:`~.Product` to search for in invoices
        """
        items = self.client.order_items.by_product(product)
        return self.from_order_items(items)

    def by_sku(self, sku: str) -> Optional[Invoice | List[Invoice]]:
        """Search for :class:`~.Invoice` s by product sku

        .. note:: Like :meth:`.OrderItemSearch.by_sku`, the sku will need to be an exact
           match to the sku of a simple product, including a custom option if applicable

           * Use :meth:`~.InvoiceSearch.by_product` or :meth:`~.InvoiceSearch.by_product_id`
             to find orders containing any of the :attr:`~.option_skus` and/or all
             :attr:`~.children` of a configurable product

        :param sku: the exact product sku to search for in invoices
        """
        items = self.client.order_items.by_sku(sku)
        return self.from_order_items(items)

    def by_product_id(self, product_id: Union[int, str]) -> Optional[Invoice | List[Invoice]]:
        """Search for :class:`~.Invoice` s by ``product_id``

        :param product_id: the ``id`` (``product_id``) of the product to search for in invoices
        """
        items = self.client.order_items.by_product_id(product_id)
        return self.from_order_items(items)

    def by_category_id(self, category_id: Union[int, str], search_subcategories: bool = False) -> Optional[Invoice | List[Invoice]]:
        """Search for :class:`~.Invoice` s by ``category_id``

        :param category_id: id of the category to search for in orders
        :param search_subcategories: if ``True``, also searches for orders from :attr:`~.all_subcategories`
        :returns: any :class:`~.Invoice` containing a :class:`~.Product` in the corresponding :class:`~.Category`
        """
        items = self.client.order_items.by_category_id(category_id, search_subcategories)
        return self.from_order_items(items)

    def by_category(self, category: Category, search_subcategories: bool = False) -> Optional[Invoice | List[Invoice]]:
        """Search for :class:`~.Invoice` s that contain any of the category's :attr:`~.Category.products`

        :param category: the :class:`~.Category` to use in the search
        :param search_subcategories: if ``True``, also searches for orders from :attr:`~.all_subcategories`
        :returns: any :class:`~.Invoice` that contains a product in the provided category
        """
        items = self.client.order_items.by_category(category, search_subcategories)
        return self.from_order_items(items)

    def by_skulist(self, skulist: Union[str, Iterable[str]]) -> Optional[Invoice | List[Invoice]]:
        """Search for :class:`~.Invoice` s using a list or comma separated string of product SKUs

        :param skulist: an iterable or comma separated string of product SKUs
        """
        items = self.client.order_items.by_skulist(skulist)
        return self.from_order_items(items)

    def from_order_items(self, items: Optional[OrderItem | List[OrderItem]]) -> Optional[Invoice, List[Invoice]]:
        """Retrieve unique :class:`~.Invoice` objects from :class:`~.OrderItem` entries using a single request

        .. tip:: Since there is no ``invoices/items`` endpoint, to search for invoices we must first do an
           :class:`OrderItemSearch`, then retrieve the ``order_ids`` and search :meth:`~.by_order_id`

        :param items: an individual/list of order items
        """
        if items is None:
            return self.client.logger.info(
                'No matching invoices for this search query'
            )
        if isinstance(items, list):
            order_ids = set(item.order_id for item in items)
            return self.by_list('order_id', order_ids)
        else:
            return self.by_order_id(items.order_id)  # Single OrderItem


class ProductSearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``products`` endpoint"""

    def __init__(self, client: Client):
        """Initialize a :class:`ProductSearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='products',
            client=client,
            model=Product
        )

    @property
    def attributes(self) -> ProductAttributeSearch:
        """Alternate way to access the SearchQuery for :class:`~.ProductAttribute` data"""
        return ProductAttributeSearch(self.client)

    def by_id(self, item_id: Union[int, str]) -> Optional[Product]:
        """Retrieve a :class:`~.Product` by ``product_id``

        .. note:: Response data from the ``products`` endpoint only has an ``id`` field, but
           all other endpoints that return data about products will use ``product_id``

        :param item_id: the ``id`` (``product_id``) of the product
        """
        return self.add_criteria(
            field='entity_id',  # Product has no "entity_id" field in API responses
            value=item_id  # But to search by the "id" field, must use "entity_id"
        ).execute()

    def by_sku(self, sku) -> Optional[Product]:
        """Retrieve a :class:`~.Product` by ``sku``

        :param sku: the product sku
        """
        return super().by_id(Model.encode(sku))

    def by_skulist(self, skulist: Union[str, Iterable[str]]) -> Optional[Product | List[Product]]:
        """Search for :class:`~.Product`s using a list or comma separated string of SKUs

        :param skulist: an iterable or comma separated string of SKUs
        """
        if not isinstance(skulist, Iterable):
            raise TypeError(f'`skulist` must be an iterable or comma-separated string of SKUs')
        if isinstance(skulist, str):
            skulist = skulist.split(',')

        skus = map(Model.encode, skulist)
        return self.by_list('sku', skus)

    def by_category(self, category: Category, search_subcategories: bool = False) -> Optional[Product | List[Product]]:
        """Search for :class:`~.Product` s in a :class:`~.Category`

        :param category: the :class:`~.Category` to retrieve products from
        :param search_subcategories: if ``True``, also retrieves products from :attr:`~.all_subcategories`
        """
        if not isinstance(category, Category):
            raise TypeError(f'`category` must be of type {Category}')

        if search_subcategories:
            category_ids = [category.id] + category.all_subcategory_ids
            return self.by_list('category_id', category_ids)
        else:
            return self.add_criteria('category_id', category.id).execute()

    def by_category_id(self, category_id: Union[int, str], search_subcategories: bool = False) -> Optional[Product | List[Product]]:
        """Search for :class:`~.Product` s by ``category_id``

        :param category_id: the id of the :class:`~.Category` to retrieve products from
        :param search_subcategories: if ``True``, also retrieves products from :attr:`~.all_subcategories`
        """
        if search_subcategories:
            if category := self.client.categories.by_id(category_id):
                return self.by_category(category, search_subcategories)
            return None
        else:
            return self.add_criteria('category_id', category_id).execute()

    def get_stock(self, sku) -> Optional[int]:
        """Retrieve the :attr:`~.stock` of a product by sku

        :param sku: the product sku
        """
        if product := self.by_sku(sku):
            return product.stock


class ProductAttributeSearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``products/attributes`` endpoint"""

    def __init__(self, client: Client):
        """Initialize a :class:`ProductAttributeSearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='products/attributes',
            client=client,
            model=ProductAttribute
        )

    def get_all(self) -> Optional[List[ProductAttribute]]:
        """Retrieve a list of all :class:`~.ProductAttribute`s"""
        return self.add_criteria('position', 0, 'gteq').execute()

    def by_code(self, attribute_code: str) -> Optional[ProductAttribute]:
        """Retrieve a :class:`~.ProductAttribute` by its attribute code

        :param attribute_code: the code of the :class:`~.ProductAttribute`
        """
        return self.by_id(attribute_code)

    def get_types(self) -> Optional[List[APIResponse]]:
        """Retrieve a list of all available :class:`~.ProductAttribute` types"""
        return self.client.search(f'{self.endpoint}/types').execute()


class CategorySearch(SearchQuery):

    """:class:`SearchQuery` subclass for the ``categories`` endpoint"""

    def __init__(self, client: Client):
        """Initialize a :class:`CategorySearch`

        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            endpoint='categories',
            client=client,
            model=Category
        )

    def by_id(self, item_id: Union[int, str]) -> Optional[Category]:
        self.query += f'rootCategoryId={item_id}'
        return self.execute()

    def by_list(self, field: str, values: Iterable) -> Optional[Category, List[Category]]:
        self.query = self.query.replace('categories', 'categories/list')
        return super().by_list(field, values)

    def get_root(self) -> Category:
        """Retrieve the top level/default :class:`~.Category` (every other category is a subcategory)"""
        return self.execute()

    def get_all(self) -> List[Category]:
        """Retrieve a list of all categories"""
        self.query = self.query.replace('categories', 'categories/list') + 'searchCriteria[currentPage]=1'
        return self.execute()

    def by_name(self, name: str, exact: bool = True) -> Optional[Category | List[Category]]:
        """Search for a :class:`~.Category` by name

        :param name: the category name to search for
        :param exact: whether the name should be an exact match
        """
        self.query = self.query.replace('categories', 'categories/list')
        if exact:
            return self.add_criteria('name', name).execute()
        else:
            return self.add_criteria('name', f'%25{name}%25', 'like').execute()



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, List, Union
from functools import cached_property
from . import Model
import copy

if TYPE_CHECKING:
    from magento import Client
    from . import Product, Invoice


class Order(Model):

    """Wrapper for the ``orders`` endpoint"""

    DOCUMENTATION = 'https://adobe-commerce.redoc.ly/2.3.7-admin/tag/orders'
    IDENTIFIER = 'entity_id'

    def __init__(self, data: dict, client: Client):
        """Initialize an Order object using an API response from the ``orders`` endpoint

        :param data: API response from the ``orders`` endpoint
        :param client: an initialized :class:`~.Client` object
        """
        super().__init__(
            data=data,
            client=client,
            endpoint='orders',
            private_keys=True
        )

    def __repr__(self):
        return f'<Magento Order: #{self.number} placed on {self.created_at}>'

    @property
    def excluded_keys(self) -> List[str]:
        return ['items', 'payment']

    @property
    def id(self) -> int:
        """Alias for ``entity_id``"""
        return getattr(self, 'entity_id', None)

    @property
    def number(self) -> str:
        """Alias for ``increment_id``"""
        return getattr(self, 'increment_id', None)

    @cached_property
    def items(self) -> List[OrderItem]:
        """The ordered items, returned as a list of :class:`OrderItem` objects

        .. note:: When a configurable :class:`~.Product` is ordered, the API returns data
           for both the configurable and simple product

           * The :class:`OrderItem` is initialized using the configurable product data, since
             the simple product data is incomplete
           * The :attr:`~.OrderItem.product` and :attr:`~.OrderItem.product_id` will still
             match the simple product though

           If both entries are needed, the unparsed response is in the :attr:`~.data` dict
        """
        return [OrderItem(item, order=self) for item in self.__items if item.get('parent_item') is None]

    @cached_property
    def item_ids(self) -> List[int]:
        """The ``item_id`` s of the ordered :attr:`~.items`"""
        return [item.item_id for item in self.items]

    @cached_property
    def products(self) -> List[Product]:
        """The ordered :attr:`~items`, returned as their corresponding :class:`~.Product` objects"""
        return [item.product for item in self.items]

    def get_invoice(self) -> Invoice:
        """Retrieve the :class:`~.Invoice` of the Order"""
        return self.client.invoices.by_order(self)

    @property
    def shipping_address(self) -> dict:
        """Shipping details, from ``extension_attributes.shipping_assignments``"""
        return self.extension_attributes.get(
            'shipping_assignments', [{}])[0].get(
            'shipping', {}).get(
            'address', {})

    @property
    def bill_to(self) -> dict:
        """Condensed version of the ``billing_address`` dict"""
        if hasattr(self, 'billing_address'):
            return {
                'firstname': self.billing_address.get('firstname', ''),
                'lastname': self.billing_address.get('lastname', ''),
                'email': self.billing_address.get('email', ''),
                'address': self.bill_to_address
            }

    @property
    def bill_to_address(self) -> str:
        """The billing address, parsed into a single string"""
        return self._build_address('billing')

    @property
    def ship_to(self) -> dict:
        """Condensed version of the :attr:`~.shipping_address` dict"""
        return {
            'firstname': self.shipping_address.get('firstname', ''),
            'lastname': self.shipping_address.get('lastname', ''),
            'email': self.shipping_address.get('email', ''),
            'address': self.ship_to_address
        }

    @property
    def ship_to_address(self) -> str:
        """The shipping address, parsed into a single string"""
        return self._build_address('shipping')

    def _build_address(self, address_type: str) -> str:
        """Parses an address dict into a single string

        :param address_type: the address to parse; either ``shipping`` or ``billing``
        """
        address_dict = getattr(self, f'{address_type}_address')
        address = ' '.join(address_dict.get('street')) + ', '
        for field in ('city', 'region_code', 'postcode', 'country_id'):
            if value := address_dict.get(field):
                address += value.replace('None', '') + ', '
        return address.strip(', ')

    @cached_property
    def payment(self) -> dict:
        """Payment data"""
        data = copy.deepcopy(self.__payment)
        if additional_info := self.extension_attributes.get('payment_additional_info'):
            data.pop('additional_information', None)
            data.update(self.unpack_attributes(additional_info, key='key'))
        return data

    @property
    def net_tax(self) -> float:
        """Final tax amount, with refunds and cancellations taken into account"""
        return self.base_tax_amount - getattr(self, 'base_tax_refunded', 0) - getattr(self, 'base_tax_canceled', 0)

    @property
    def net_total(self) -> float:
        """Final Order value, with refunds and cancellations taken into account"""
        return self.base_grand_total - getattr(self, 'base_total_refunded', 0) - getattr(self, 'base_total_canceled', 0)

    @property
    def item_refunds(self) -> float:
        """Total amount refunded for items; excludes shipping and adjustment refunds/fees"""
        return sum(item.net_refund for item in self.items)

    @cached_property
    def total_qty_invoiced(self) -> int:
        """Total number of units invoiced"""
        return sum(item.qty_invoiced for item in self.items)

    @cached_property
    def total_qty_shipped(self) -> int:
        """Total number of units shipped"""
        return sum(item.qty_shipped for item in self.items)

    @cached_property
    def total_qty_refunded(self) -> int:
        """Total number of units refunded"""
        return sum(item.qty_refunded for item in self.items)

    @cached_property
    def total_qty_canceled(self) -> int:
        """Total number of units canceled"""
        return sum(item.qty_canceled for item in self.items)

    @cached_property
    def total_qty_outstanding(self) -> int:
        """Total number of units that haven't been shipped/fulfilled yet"""
        return sum(item.qty_outstanding for item in self.items)

    @cached_property
    def net_qty_ordered(self) -> int:
        """Total number of units ordered, after accounting for refunds and cancellations"""
        return sum(item.net_qty_ordered for item in self.items)


class OrderItem(Model):

    """Wrapper for the ``order/items`` endpoint"""

    DOCUMENTATION = "https://adobe-commerce.redoc.ly/2.3.7-admin/tag/ordersitems"
    IDENTIFIER = 'item_id'

    def __init__(self, item: dict, client: Optional[Client] = None, order: Optional[Order] = None):
        """Initialize an OrderItem using an API response from the ``orders/items`` endpoint

        .. note:: Initialization requires either a :class:`~.Client` or :class:`Order` object

        :param item: API response from the ``orders/items`` endpoint
        :param order: the :class:`Order` that this is an item of
        :param client: the :class:`~.Client` to use (if not initializing with an Order)
        :raise ValueError: if both the ``order`` and ``client`` aren't provided
        """
        if client is None:
            if order is None:
                raise ValueError('An Order or Client object must be provided')
            if not isinstance(order, Order):
                raise TypeError(f'`order` must be of type {Order}')

        super().__init__(
            data=item,
            client=client if client else order.client,
            endpoint='orders/items'
        )
        self.tax = item.get('base_tax_amount', item.get('tax_amount', 0))
        self.refund = item.get('base_amount_refunded', item.get('amount_refunded', 0))
        self.tax_refunded = item.get('base_tax_refunded', item.get('tax_refunded', 0))
        self.line_total = item.get('base_row_total_incl_tax', item.get('row_total_incl_tax', 0))
        self._order = order

    def __repr__(self):
        return f"<OrderItem ({self.sku})> from Order ID: {self.order_id}>"

    @property
    def excluded_keys(self) -> List[str]:
        return ['product_id']

    @cached_property
    def order(self) -> Order:
        """The corresponding :class:`Order`"""
        if self._order is None:
            return self.client.orders.by_id(self.order_id)
        return self._order

    @cached_property
    def product(self) -> Product:
        """The item's corresponding :class:`~.Product`

        .. note:: **If the ordered item:**

           * Is a configurable product - the child simple product is returned
           * Has custom options - the base product is returned
        """
        if self.product_type != 'configurable':
            return self.client.products.by_id(self.product_id)

        if not self.extension_attributes.get('custom_options'):
            return self.client.products.by_sku(self.sku)

        # Configurable + Custom Options -> configurable product id & unsearchable option sku
        for item in self.order.data['items']:  # Get simple product id from response data
            if item.get('parent_item_id') == self.item_id:
                return self.client.products.by_id(item['product_id'])

    @cached_property
    def product_id(self) -> int:
        """Id of the corresponding simple :class:`~.Product`"""
        return self.__product_id if self.product_type != 'configurable' else self.product.id

    @cached_property
    def extension_attributes(self) -> dict:
        return getattr(self, 'product_option', {}).get('extension_attributes', {})

    @cached_property
    def qty_outstanding(self) -> int:
        """Number of units that haven't been shipped/fulfilled yet"""
        return self.net_qty_ordered - self.qty_shipped

    @cached_property
    def net_qty_ordered(self) -> int:
        """Number of units ordered, after accounting for refunds and cancellations"""
        return self.qty_ordered - self.qty_refunded - self.qty_canceled

    @cached_property
    def net_tax(self) -> float:
        """Tax amount after accounting for refunds and cancellations"""
        return self.tax - self.tax_refunded - getattr(self, 'tax_canceled', 0)

    @cached_property
    def net_total(self) -> float:
        """Row total (incl. tax) after accounting for refunds and cancellations"""
        return self.line_total - self.net_refund - self.total_canceled

    @cached_property
    def net_refund(self) -> float:
        """Refund amount after accounting for tax and discount refunds"""
        return self.refund + self.tax_refunded - getattr(self, 'discount_refunded', 0)

    @cached_property
    def total_canceled(self) -> float:
        """Cancelled amount; note that partial cancellation is not possible"""
        if self.qty_canceled != 0:
            return self.line_total
        return 0



from __future__ import annotations
from typing import Union, Optional, TYPE_CHECKING
import requests

if TYPE_CHECKING:
    from . import Client


class MagentoError(Exception):

    """Base exception class for error responses returned by the Magento API

    :cvar DEFAULT_MSG: default exception message to use if a message isn't provided
    """

    DEFAULT_MSG = 'An error occurred while processing the request.'

    def __init__(self, client: Client, msg: Optional[str] = None, response: Optional[requests.Response] = None):
        """Log and raise a MagentoError

        :param client: an initialized :class:`~.Client` object
        :param msg: optional exception message; prepended to the error message of the response (if provided).
        :param response: optional response to :meth:`parse` an error message from
        """
        self.message = msg if msg else self.DEFAULT_MSG
        self.logger = client.logger

        if response is not None:
            self.message += '\n' + self.parse(response)

        self.logger.error(self.message)
        super().__init__(self.message)

    @staticmethod
    def parse(response: Union[requests.Response, dict]) -> str:
        """Parses the error message from the response, including message parameters when available

        :param response: a bad response returned by the Magento API
        """
        if isinstance(response, requests.Response):
            response = response.json()

        message = response.get('message', '')
        params = response.get('parameters')
        errors = response.get('errors')

        if message:
            message = f'Message: "{message}"'

        if errors:
            for error in errors:
                err_msg = error['message']
                err_params = error.get('parameters')

                if err_params:
                    for param in err_params:
                        err_msg = err_msg.replace(f'%{param}', f'{err_params[param]}')

                message += '\n' + err_msg

        if params:
            if isinstance(params, dict):
                for param in params:  # Same format as error params
                    message = message.replace(f'%{param}', f'{params[param]}')

            elif isinstance(params, list):  # List of param values
                for i, param in enumerate(params, 1):  # Message has params as %{index}, starting from 1
                    message = message.replace(f'%{i}', param)

        return message


class AuthenticationError(MagentoError):

    """Exception class for errors when trying to :meth:`~.authenticate` a :class:`~.Client`"""

    DEFAULT_MSG = 'Failed to authenticate credentials.'

    def __init__(self, client: Client, msg: Optional[str] = None, response: Optional[requests.Response] = None):
        super().__init__(client, msg, response)



from __future__ import annotations
import json
import pickle
import requests
from functools import cached_property
from typing import Optional, Dict, List
from .utils import MagentoLogger, get_agent
from .models import APIResponse, ProductAttribute
from .search import SearchQuery, OrderSearch, ProductSearch, InvoiceSearch, CategorySearch, ProductAttributeSearch, OrderItemSearch
from .exceptions import AuthenticationError, MagentoError


class Client:

    """The class that handles all interaction with the API"""

    def __init__(
            self,
            domain: str,
            username: str,
            password: str,
            scope: Optional[str] = '',
            user_agent: Optional[str] = None,
            token: Optional[str] = None,
            log_level: str = 'INFO',
            login: bool = True,
            **kwargs
    ):
        """Initialize a Client

        .. admonition:: Important!
           :class: important-af

           The Magento account you use to log in must be assigned a **User Role** that has
           the appropriate API resources included in its **Resource Access** settings

           This can be verified in Magento Admin by going to::

            System -> Permissions -> User Roles -> {Role} -> Role Resources -> Resource Access

           and ensuring that ``Sales``, ``Catalog``, ``Customers``, and any other desired resources are included


        :param domain: domain name of the Magento store (ex. ``domain.com``)
        :param username: username of the Magento Admin account
        :param password: password of the Magento Admin account
        :param scope: the store view scope to :meth:`~search` and make requests on
        :param user_agent: the user agent to use in requests
        :param token: an existing access token
        :param log_level: the logging level for logging to stdout
        :param login: if ``True``, calls :meth:`~.authenticate` upon initialization
        :param kwargs: see below
        ...

        :Optional Keyword Arguments:
            * **log_file** (``str``) – log file to use for the client's :attr:`logger`
            * **log_requests** (``bool``) - if ``True``, the logs from :mod:`requests`
              will be added to the client's ``log_file``

        """
        #: The base API URL
        self.BASE_URL: str = f'https://www.{domain}/rest/V1/'
        #: The user credentials
        self.USER_CREDENTIALS: Dict[str, str] = {
            'username': username,
            'password': password
        }
        #: The API access token
        self.ACCESS_TOKEN: str = token
        #: The Magento store domain
        self.domain: str = domain
        #: The store view code to request/update data on
        self.scope: str = scope
        #: The user agent to use in requests
        self.user_agent: str = user_agent if user_agent else get_agent()
        #: The :class:`~.MagentoLogger` for the domain/username combination
        self.logger: MagentoLogger = self.get_logger(
            stdout_level=log_level,
            log_file=kwargs.get('log_file', None),
            log_requests=kwargs.get('log_requests', True)
        )
        #: An initialized :class:`Store` object
        self.store: Store = Store(self)

        if login:
            self.authenticate()

    @classmethod
    def new(cls) -> Client:
        """Prompts for input to log in to the Magento API"""
        return cls(
            input('Domain: '),
            input('Username: '),
            input('Password: ')
        )

    @classmethod
    def load(cls, pickle_bytes: bytes) -> Client:
        """Initialize a :class:`~.Client` using a pickle bytestring from :meth:`~.to_pickle`"""
        return pickle.loads(pickle_bytes)

    @classmethod
    def from_json(cls, json_str: str) -> Client:
        """Initialize a :class:`~.Client` from a JSON string of settings"""
        return cls.from_dict(json.loads(json_str))

    @classmethod
    def from_dict(cls, d: dict) -> Client:
        """Initialize a :class:`~.Client` from a dictionary of settings"""
        return cls(**d)

    def url_for(self, endpoint: str, scope: str = None) -> str:
        """Returns the appropriate request url for the given API endpoint and store scope

        .. admonition:: Example
            :class: example

            ::

             # Generate the url for credit memo with id 7
             >> api=Client("domain.com", "user", "password")
             >> api.url_for('creditmemo/7')
             "https://domain.com/rest/V1/creditmemo/7"

             # Generate the same url on the "en" store view
             >> api.url_for('creditmemo/7', scope='en')
             "https://domain.com/rest/en/V1/creditmemo/7"

        :param endpoint: the API endpoint
        :param scope: the scope to generate the url for; uses the :attr:`.Client.scope` if not provided
        """
        if not scope:
            if self.scope and scope is None:
                scope = self.scope
            else:
                return self.BASE_URL + endpoint
        return self.BASE_URL.replace('/V1', f'/{scope}/V1') + endpoint

    def search(self, endpoint: str) -> SearchQuery:
        """Initializes and returns a :class:`~.SearchQuery` corresponding to the specified endpoint

        .. note:: Several endpoints have predefined :class:`~.SearchQuery` and :class:`~.Model` subclasses

           If a subclass hasn't been defined for the ``endpoint`` yet, a general :class:`~.SearchQuery`
           will be returned, which wraps the :attr:`~.result` in :class:`~.APIResponse` objects

        :param endpoint: a valid Magento API search endpoint
        """
        if endpoint.lower() == 'orders':
            return self.orders
        if endpoint.lower() == 'orders/items':
            return self.order_items
        if endpoint.lower() == 'invoices':
            return self.invoices
        if endpoint.lower() == 'categories':
            return self.categories
        if endpoint.lower() == 'products':
            return self.products
        if endpoint.lower() == 'products/attributes':
            return self.product_attributes
        # Any other endpoint is queried with a general SearchQuery object
        return SearchQuery(endpoint=endpoint, client=self)

    @property
    def orders(self) -> OrderSearch:
        """Initializes an :class:`~.OrderSearch`"""
        return OrderSearch(self)

    @property
    def order_items(self) -> OrderItemSearch:
        """Initializes an :class:`~.OrderItemSearch`"""
        return OrderItemSearch(self)

    @property
    def invoices(self) -> InvoiceSearch:
        """Initializes an :class:`~.InvoiceSearch`"""
        return InvoiceSearch(self)

    @property
    def categories(self) -> CategorySearch:
        """Initializes a :class:`~.CategorySearch`"""
        return CategorySearch(self)

    @property
    def products(self) -> ProductSearch:
        """Initializes a :class:`~.ProductSearch`"""
        return ProductSearch(self)

    @property
    def product_attributes(self) -> ProductAttributeSearch:
        """Initializes a :class:`~.ProductAttributeSearch`"""
        return ProductAttributeSearch(self)

    def get(self, url: str) -> requests.Response:
        """Sends an authorized ``GET`` request

        :param url: the URL to make the request on
        """
        return self.request('GET', url)

    def post(self, url: str, payload: dict) -> requests.Response:
        """Sends an authorized ``POST`` request

        :param url: the URL to make the request on
        :param payload: the JSON payload for the request
        """
        return self.request('POST', url, payload)

    def put(self, url: str, payload: dict) -> requests.Response:
        """Sends an authorized ``PUT`` request

        :param url: the URL to make the request on
        :param payload: the JSON payload for the request
        """
        return self.request('PUT', url, payload)

    def delete(self, url: str) -> requests.Response:
        """Sends an authorized ``DELETE`` request

        :param url: the URL to make the request on
        """
        return self.request('DELETE', url)

    def authenticate(self) -> bool:
        """Authenticates the :attr:`~.USER_CREDENTIALS` and retrieves an access token
        """
        endpoint = self.url_for('integration/admin/token')
        payload = self.USER_CREDENTIALS
        headers = {
            'Content-Type': 'application/json',
            'User-Agent': self.user_agent
        }
        self.logger.info(f'Authenticating {payload["username"]} on {self.domain}...')
        response = requests.post(
            url=endpoint,
            json=payload,
            headers=headers
        )
        if response.ok:
            self.ACCESS_TOKEN = response.json()
        else:
            raise AuthenticationError(self, response=response)

        self.logger.debug('Validating token...')
        try:
            self.validate()
        except AuthenticationError as e:
            raise AuthenticationError(self, msg='Token validation failed') from e

        self.logger.info('Logged in to {}'.format(payload["username"]))
        return True

    def validate(self) -> bool:
        """Validates the :attr:`~.token` by sending an authorized request to a standard API endpoint
        """
        response = self.get(self.url_for('store/websites'))
        if response.status_code == 200:
            self.logger.debug("Token validated for {} on {}".format(
                self.USER_CREDENTIALS['username'], self.domain)
            )
            return True
        else:
            msg = "Token validation failed for {} on {}".format(
                self.USER_CREDENTIALS['username'], self.domain
            )
            raise AuthenticationError(self, msg=msg, response=response)

    def request(self, method: str, url: str, payload: dict = None) -> requests.Response:
        """Sends an authorized API request. Used for all internal requests

        .. tip:: Use :meth:`get`, :meth:`post`, :meth:`put` or :meth:`delete` instead

        :param method: the request method
        :param url: the url to send the request to
        :param payload: the JSON payload for the request (if the method is ``POST`` or ``PUT``)
        """
        if method in ('GET', 'DELETE'):
            response = requests.request(method, url, headers=self.headers)
        elif method in ('POST', 'PUT'):
            if payload:
                response = requests.request(method, url, json=payload, headers=self.headers)
            else:
                raise ValueError('Must provide a non-empty payload')
        else:
            raise ValueError('Invalid request method provided')

        if response.status_code == 401:
            self.logger.debug("Attempting to re-authenticate...")
            self.authenticate()  # Will raise AuthenticationError if unsuccessful (won't recurse infinitely)
            return self.request(method, url, payload)

        if response.status_code != 200:  # All other responses are returned; errors are handled by methods
            self.logger.error("Request to {} failed with status code {}.\n{message}".format(
                url, response.status_code, message=MagentoError.parse(response))
            )

        return response

    def get_logger(self, log_file: str = None, stdout_level: str = 'INFO', log_requests: bool = True) -> MagentoLogger:
        """Retrieve a MagentoLogger for the current username/domain combination. Log files are DEBUG.

        :param log_file: the file to log to
        :param stdout_level: the logging level for stdout logging
        :param log_requests: if ``True``, adds the :class:`~.FileHandler` to the :mod:`~.urllib3.connectionpool` logger
        """
        logger_name = MagentoLogger.CLIENT_LOG_NAME.format(
            domain=self.domain.split('.')[0],
            username=self.USER_CREDENTIALS['username']
        )   # Example:``domain_username``
        return MagentoLogger(
            name=logger_name,
            log_file=log_file,
            stdout_level=stdout_level,
            log_requests=log_requests
        )

    @property
    def headers(self) -> dict:
        """Authorization headers for API requests

        Automatically generates a :attr:`token` if needed
        """
        return {
            'Authorization': f'Bearer {self.token}',
            'User-Agent': self.user_agent
        }

    @property
    def token(self) -> str:
        """Returns or generates an :attr:`~ACCES_TOKEN`"""
        if not self.ACCESS_TOKEN:
            self.authenticate()
        return self.ACCESS_TOKEN

    def to_pickle(self, validate: bool = False) -> bytes:
        """Serializes the Client to a pickle bytestring

        :param validate: if ``True``, validates the :attr:`token`/:attr:`USER_CREDENTIALS` before serializing
        """
        if validate:
            self.validate()
        return pickle.dumps(self)

    def to_json(self, validate: bool = False) -> str:
        """Serializes the Client to a JSON string

        :param validate: if ``True``, validates the :attr:`token`/:attr:`USER_CREDENTIALS` before serializing
        """
        data = self.to_dict(validate)
        return json.dumps(data)

    def to_dict(self, validate: bool = False) -> Dict[str, str]:
        """Serializes the Client to a dictionary

        :param validate: if ``True``, validates the :attr:`token`/:attr:`USER_CREDENTIALS` before serializing
        """
        if validate:
            self.validate()
        data = {
            'domain': self.domain,
            'username': self.USER_CREDENTIALS['username'],
            'password': self.USER_CREDENTIALS['password'],
            'scope': self.scope,
            'user_agent': self.user_agent,
            'token': self.token,
            'log_level': self.logger.logger.level,
            'log_file': self.logger.log_file
        }
        return data

    def view_config(self):
        """Prints the Client configuration settings"""
        for k, v in self.to_dict().items():
            print(f'{k} : {v}')


class Store:

    """Class containing store configurations and cached attribute lists"""

    def __init__(self, client: Client):
        """Initialize a Store object

        :param client: an initialized :class:`~.Client` object
        """
        self.client = client

    @property
    def is_single_store(self) -> bool:
        """Whether the store has a single store view (``default``) or multiple store views"""
        return len(self.configs) == 1

    @property
    def active(self) -> APIResponse:
        """Returns the store config corresponding to the current :attr:`~.Client.scope` of the :class:`Client`"""
        store_code = 'default' if self.client.scope in ('', 'all') else self.client.scope
        for store in self.configs:
            if store.code == store_code:
                return store

    @cached_property
    def configs(self) -> Optional[APIResponse | List[APIResponse]]:
        """Returns a list of all store configurations"""
        return self.client.search('store/storeConfigs').execute()

    @cached_property
    def views(self) -> Optional[APIResponse | List[APIResponse]]:
        """Returns a list of all store views"""
        return self.client.search('store/storeViews').execute()

    @cached_property
    def all_product_attributes(self) -> List[ProductAttribute]:
        """A cached list of all product attributes"""
        return self.client.product_attributes.get_all()

    @cached_property
    def store_view_product_attributes(self) -> List[ProductAttribute]:
        """A cached list of all product attributes with the ``Store View`` scope"""
        return [attr for attr in self.all_product_attributes if attr.scope == 'store']

    @cached_property
    def website_product_attributes(self) -> List[ProductAttribute]:
        """A cached list of all product attributes with the ``Web Site`` scope"""
        return [attr for attr in self.all_product_attributes if attr.scope == 'website']

    @cached_property
    def global_product_attributes(self) -> List[ProductAttribute]:
        """A cached list of all product attributes with the ``Global`` scope"""
        return [attr for attr in self.all_product_attributes if attr.scope == 'global']

    @cached_property
    def website_attribute_codes(self) -> List[str]:
        """The attribute codes of the :attr:`~.website_product_attributes`"""
        return [attr.attribute_code for attr in self.website_product_attributes]

    def filter_website_attrs(self, attribute_data: dict) -> dict:
        """Filters a product attribute dict and returns a new one that contains only the website scope attributes

        Website scoped attributes must be updated on the admin by making a second request on the ``all`` scope

        * This method is called by :meth:`~.Product.update_attributes` and :meth:`~.Product.update_custom_attributes`
          to see if the second request is needed

        .. admonition:: **Example**
           :class: example

           The ``price`` attribute is ``Website`` scope and the ``meta_title`` attribute is ``Store View`` scope

           ::

            >> attribute_data = {'price': 12, 'meta_title': 'My Product'}
            >> store.filter_website_attrs(attribute_data)
            {'price': 12}

        :param attribute_data: a dict of product attributes
        """
        return {k: v for k, v in attribute_data.items() if k in self.website_attribute_codes}

    def refresh(self) -> bool:
        """Clears all cached properties"""
        cached = ('configs', 'views', 'all_product_attributes', 'store_view_product_attributes',
                  'website_product_attributes', 'global_product_attributes', 'website_attribute_codes')
        for key in cached:
            self.__dict__.pop(key, None)
        return True
